#!/usr/bin/tcl
#
# quote.tcl v2.4b
#
#  By: [KM] - km@grogg.org
#
# About:
#
#  This is a small (but growing) script that should be used for storing
#  quotes in. The idea is to capture funny oneliners and other interesting
#  things that people say. But really it could be anything from urls to
#  jokes or whatever you can think of.
#  When the quote is stored you can use a command to display a random quote
#  or pick one by referencing its number. Other small utilities to show
#  top quoters and output to the web is included. Whenever I can think of
#  something that would be cool to add, i add it (See Features and Todo). 
#
#  If you feel like it, join #stockholm on IRCnet, home of tori and
#  the quote script.
#
#  For new versions of this script and more, check out
#  http://www.grogg.org/tori/
#
#
# History:
#
#  1.0 - First version, basic functionality and quite full of bugs.
#
#  1.1 - Most bugs fixed and !quote and !putquote fully implemented
#
#  1.2 - New function !qstat, which lists number of quotes stored.
#        The "quote chars" bug fixed.
#        Added Html output of quote database via a (bourne) shell script.
#
#  1.3 - New command !lastquote, which will display the last quote entered.
#        New command !topq (uses external perl script), which will display the
#        most frequent users of !putquote. Some rearrangement of code.
#
#  1.4 - New functionality to !quote command. Users are able to request
#        a quote in the database by its number (line in the db). Output of a
#        quote also includes output of its number.
#        New command !qfind which will show a random quote from all hits (case
#        insensitive (requested)). Uses external perl script qfind.pl.
#        A help command is added, !qhelp.
#        Command names are changed to be more consistent (beginning with q).
#
#  1.5 - Added average quote length to !qstat and to html output.
#        Added new functionality to !qlast. !qlast <nick> shows the last quote
#        entered by <nick> (only quick fix done, should be remade ASAP).
#        Minor code cleanup. Remade the quote2html script with much better
#        looking html instead of the old <pre></pre> style. Fixed so that
#        !qfind shows which hit of how many matches the displayed quote is.
#        !qput now returns length of quote stored.
#
#  1.6 - Added !qlist command which is exactly as !qfind except that it
#        returns a full index of all matches instead of a random quote.
#        Added a !qdate command, which will return a full index of all quotes
#        added that date.
#        Fixed a millenium bug. ;)
#        Joins and major rewrites of the perl backend scripts.
#
#  1.7 - Small bugfixes. Finally fixed the layout in the HTML output from
#        quote2html.sh. Added a small script that puts the quote db into
#        a mysql table and 2 php pages. One that outputs a random quote
#        and one that searches and returns all matching quotes. Small code
#        cleanup. Minor calculation and shell interaction fixes.
#
#  1.8 - Since I upgraded my eggdrop from 1.1.5 to 1.6.10 (about time? :D)
#        this script is also rewritten to work with 1.6.x. Changed flag
#        needed for !qput to Q. Added an info command.
#
#  2.0 - The script now stores and retrieves all quotes directly from a mysql
#        database using mysqltcl and DBI::MySQL.
#        Changed the !quote command so that when the input isnt an integer
#        a !qfind search is made instead. The !qfind command is still there so
#        that you can search for numbers.
#
#  2.1 - Changed !quote so that when a negative number is entered it shows the
#        quote with that number subtracted from the last quote.
#        Added an ignore channel variable. Fixed a bug in qfind from 2.0.
#        Added a TIMESTAMP column so that the exact time will be stored.
#        !qinfo now either gives info about a string or details about a quote.
#        Redesigned the DB (not compatible with 2.0 DB).
#
#  2.2 - Finally got rid of the external qtop.pl. Now its all tcl!
#        Changed !qstat a little. Some code cleanup.
#        Added some bold text to clarify output (optional).
#        Changed all commands to be able to print more than one line of
#        output (se max_output setting to control how many).
#
#  2.3 - Added color to output (optional). Changed output style a little.
#        You now have to specify which channels the quotescript should be
#        active in. Removed ignore_channel for obvious reasons :)
#        Added a mysqldb variable so you can call the db whatever you want.
#        Add a -nick parameter to qlist: !qlist -nick NICK to display all
#        quotes added by NICK.
#        Add a -nick parameter to qinfo: !qinfo -nick NICK to info about NICK.
#        Resolved a security issue with SQL injection in SELECT statements.
#
#  2.4 - Fixed a minor formatting anomality.
#        Added -nick parameter to !q/!qfind.
#        Changed the !qdel command. Only the user who last !qputed a quote
#        can delete that quote.
#        Fixed a very embarassing bug. The script couldnt handle an empty DB :D
#
#
# Features:
#
#   command     input          attr description
#
#   !qput       text           Q    Store a quote.
#   !quote|!q   [int|string]   -    Output a random quote or quote <number>.
#   !qfind      string         -    Show a random quote containing <string>.
#   !qstat                     -    See some stats about stored quotes.
#   !qlast      [nick] [-int]  -    See last quote entered (by <nick>).
#   !qtop       [int]          -    Show (default|<number>) of top quoters.
#   !qhelp      -              -    Show available commands.
#   !qlist      [-nick] string -    Show an index of quotes containing <string>.
#   !qdate      YYYY-MM-DD|    -    Show an index of all quotes from a date.
#               YYYYMMDD
#   !qinfo      [-nick]        -    Show stats about a quote or a string.
#               string|int     
#   !qdel       -              Q    Deletes last stored quote.
#
#  All quotes stored in mysql.
#
#  Example php scripts included to view and search quotes on the web.
#
#
# Installation:
#
#  1. Download and install libmysqltcl:
#     http://www.xdobry.de/mysqltcl/
#
#  2. Create the database:
#     $ mysqladmin create quotes
#
#  3. Set up a table in the database.
#     $ mysql quotes
#     > CREATE TABLE quotes (qid INT(11) NOT NULL AUTO_INCREMENT, quote TEXT,
#       name VARCHAR(9), timestamp TIMESTAMP NOT NULL, PRIMARY KEY (qid));
#
#  4. chmod 600 quote.tcl (so people cant read your mysql password).
#
#  5. Edit quote.tcl (this file) to configure.
#
#  6. Add quote.tcl to your eggdrop config and edit the settings below.
#
#  7. Issue .chattr <user> +Q for all users that should be able to use !qput.
#
# 
# Todo:
#
#  1. Implement an admin command and/or an admin php-page.
#
#  2. Develop a smack habit.
#
# 
# Dependencies:
#
#  quote.tcl	  	Main eggdrop script.
#
#  quote.php		Shows a random quote from the DB.
#  quotes.php		Shows all quotes stored in the DB.
#
#  libmysqltcl		Found at http://www.xdobry.de/mysqltcl/
#  
#  PHP compiled with MySQL support (If you want web access)
#  (PHP can be found at http://www.php.net)
#
#
### SETTINGS ##################################################################

# Active channels
set channels "#stockholm"

# Maximum number of lines to wrap output to
set max_output 3

# Bold text or not (1/0)
set bold 1

# Colored text or not (1/0)
set color 1

# Number of people to list by default when no argument is given to qtop
set qtop 5

# The homepage URL to show in the !qhelp command
set homepage "http://www.grogg.org/tori/"

# Where in the world is libmysqltcl?
set libmysqltcl "/usr/local/lib/libmysqltcl.so"

# User, password and host for Mysql connection
set mysqluser "user"
set mysqlpass "pass"
set mysqlhost "host"
set mysqldb   "quotes"


### NO EDITING BELOW THIS LINE ################################################


# Load mysql interface
load $libmysqltcl
set handle [mysqlconnect -user $mysqluser -host $mysqlhost -password $mysqlpass]

# bind commands
bind pub Q|Q !qput put_quote
bind pub -|- !q get_quote	
bind pub -|- !quote get_quote
bind pub -|- !qstat stat_quote
bind pub -|- !qlast last_quote
bind pub -|- !qtop top_quote
bind pub -|- !qhelp help_quote
bind pub -|- !qfind find_quote
bind pub -|- !qlist list_quote
bind pub -|- !qdate date_quote
bind pub -|- !qinfo info_quote
bind pub Q|Q !qdel del_quote
#bind pub A|A !qadmin admin_quote

# Am I bold or not?
if { $bold } {
  set b ""
} else {
  set b ""
}

# Am I in color or not?
if { $color } {
  set c  "\003"
  set c2 "\00314"
} else {
  set c ""
  set c2 ""
}

#------------------------------------------------------------------------------
# Small general procs
#------------------------------------------------------------------------------

# Pad time with zeroes
proc fix_time { hour min } {
  if { [string length $hour] != 2 } { set hour "0$hour" }
  if { [string length $min]  != 2 } { set min  "0$min" }
  return "$hour:$min"
}

# Strip input of bad chars
proc strip_bad { arg } {
  regsub -all -- {'} $arg {\'} arg
  regsub -all -- {\\\\+} $arg {\\} arg
  return $arg
}

# Format qid
proc fmt_qid { qid hit hits } {
  global b c c2
  if { "$hit$hits" != "" } {
    set output "$c2\[$c$b$qid$b$c2|${c}($b$hit${b}/${b}$hits${b})${c2}\]$c:"
  } else {
    set output "$c2\[$c$b$qid$b$c2\]$c:"
  }
  return $output
}

# Format by-line
proc fmt_by { name hour min day month year } {
  global b c c2
  return "${b}(${b}by $name [fix_time $hour $min] $day $month $year${b})"
}

# Formatted putchan for more than one line of output
proc fmt_putchan { chan output } {
  global max_output
  set output_length 0
  set i 0
  append output " "

  while { $output_length != -1 && $i < $max_output } {
    set output_length [string last " " $output\
                      [expr 453 - [string length $chan]]]
    putchan $chan [string range $output 0 $output_length]
    set output [string replace $output 0 $output_length]
    incr i
  }
  return 0
}

# Check if the channel is active
proc active_chan { chan } {
  global channels

  foreach active_chan [split $channels] {
    if { ![string compare -nocase $active_chan $chan] } {
      return 1
    }
  }
  return 0
}

#------------------------------------------------------------------------------
# Insert a quote in the quotedb.
#------------------------------------------------------------------------------
proc put_quote { nick uhost hand chan args } {
  global handle mysqldb b
  if { ![active_chan $chan] } { return 0 }

  set arg [strip_bad [string trim [lindex $args 0]]]
  mysqluse $handle $mysqldb

  set qid [mysqlsel $handle "SELECT MAX(qid) FROM quotes" -flatlist]
  if { $qid == "{}" } { set qid 0 }
  mysqlexec $handle "INSERT INTO quotes\
                     VALUES([incr qid], '$arg', '$hand', NULL)"

  putcmdlog "#$hand# quote $qid stored..."
  putchan $chan "$nick: Quote $b$qid$b stored at [clock format [clock seconds]\
                 -format %R] ([string length $arg] chars)..."
  return 0
}

#------------------------------------------------------------------------------
# Delete the last quote in the quotedb.
#------------------------------------------------------------------------------
proc del_quote { nick uhost hand chan args } {
  global handle mysqldb b
  if { ![active_chan $chan] } { return 0 }

  set arg [strip_bad [lindex $args 0]]
  mysqluse $handle $mysqldb 

  set arg [mysqlsel $handle "SELECT MAX(qid) FROM quotes" -flatlist]
  set db_hand [mysqlsel $handle "SELECT name FROM quotes WHERE qid = $arg"\
               -flatlist]

  if { ![string compare $hand $db_hand] } {
    mysqlexec $handle "DELETE FROM quotes WHERE qid = $arg"
    putcmdlog "#$hand# quote $arg deleted..."
    set output "$nick: Quote $b$arg$b deleted..."
  } else {
    set output "$nick: Last quote was added by $db_hand and you are $hand."
    putcmdlog "#$hand# tried to delete last quote, but was denied..."
  }
  putchan $chan $output
  return 0
}

#------------------------------------------------------------------------------
# Print a quote to channel.
#------------------------------------------------------------------------------
proc get_quote { nick uhost hand chan args } {
  global handle mysqldb
  if { ![active_chan $chan] } { return 0 }

  set arg [lindex $args 0]

  mysqluse $handle $mysqldb

  set quotes [mysqlsel $handle "SELECT MAX(qid) FROM quotes" -flatlist]

  if { $arg != "" } { 
    if { ![string is integer -strict $arg] } {
      find_quote $nick $uhost $hand $chan $arg
      return 0
    }
  } else {
    set arg [expr [rand $quotes] + 1]
  }

  if { [string index $arg 0] == "-" } { set arg [expr $quotes $arg] }

  if { $arg > $quotes || $arg < 1 } {
    set output "ERROR: Database only contains $quotes quotes."
  } else {
    mysqlsel $handle "SELECT qid,quote,name,YEAR(timestamp),\
                      LEFT(MONTHNAME(timestamp), 3),DAYOFMONTH(timestamp),\
                      HOUR(timestamp),MINUTE(timestamp) FROM quotes\
                      WHERE qid = $arg"
    mysqlmap $handle { qid quote name year month day hour min } {
      set output "[fmt_qid $qid "" ""] [string trim $quote]\
                  [fmt_by $name $hour $min $day $month $year]"
    #" (this line is for fixing a bug in vim's syntax higlighting.)
    }

    if { ![info exists output] } {
      set output "ERROR: quote $qid not found."
    }
  }
  fmt_putchan $chan $output
  return 0
}

#------------------------------------------------------------------------------
# Find a quote and print to channel.
#------------------------------------------------------------------------------
proc find_quote { nick uhost hand chan args } {
  global handle mysqldb
  if { ![active_chan $chan] } { return 0 }
     
  set arg  [strip_bad [lindex [split [lindex $args 0] " "] 0]]
  set arg2 [strip_bad [lindex [split [lindex $args 0] " "] 1]]

  mysqluse $handle $mysqldb

  if { $arg != "" } { 
    if { ![string compare -nocase $arg "-nick"] } {
      set arg $arg2
      set quote_hits [mysqlsel $handle "SELECT qid,quote,name,YEAR(timestamp),\
                      LEFT(MONTHNAME(timestamp), 3),DAYOFMONTH(timestamp),\
                      HOUR(timestamp),MINUTE(timestamp) FROM quotes\
                      WHERE name = '$arg'"]
    } else {
      set arg [strip_bad [lindex $args 0]]
      set quote_hits [mysqlsel $handle "SELECT qid,quote,name,YEAR(timestamp),\
                      LEFT(MONTHNAME(timestamp), 3),DAYOFMONTH(timestamp),\
                      HOUR(timestamp),MINUTE(timestamp) FROM quotes\
                      WHERE quote LIKE '%$arg%'"]
    }

    if { !$quote_hits } {
      set output "$nick: no match for \"$arg\""
    } else {
      set quote_num [expr [rand $quote_hits]]
      mysqlseek $handle $quote_num
      set output [mysqlnext $handle]
      set output "[fmt_qid [lindex $output 0] [incr quote_num] $quote_hits]\
                  [string trim [lindex $output 1]] [fmt_by [lindex $output 2]\
                  [lindex $output 6] [lindex $output 7] [lindex $output 5]\
                  [lindex $output 4] [lindex $output 3]]"
    }
  } else {
    set output "ERROR: you have to specify a search string"
  }
  putcmdlog "#$hand# quoting on $chan"
  fmt_putchan $chan $output
  return 0
}

#------------------------------------------------------------------------------
# Print info about a quote, a string or a nick to channel.
#------------------------------------------------------------------------------
proc info_quote { nick uhost hand chan args } {
  global handle mysqldb b
  if { ![active_chan $chan] } { return 0 }

  set arg  [strip_bad [lindex [split [lindex $args 0] " "] 0]]
  set arg2 [strip_bad [lindex [split [lindex $args 0] " "] 1]]

  mysqluse $handle $mysqldb

  if { $arg != "" } { 
    if { ![string compare -nocase $arg "-nick"] } {
      set arg $arg2
        set maxqid [mysqlsel $handle "SELECT COUNT(qid) FROM quotes" -flatlist]
        set quotes [mysqlsel $handle "SELECT qid FROM quotes WHERE\
                                      name = '$arg'" -list]
        set output "$arg has added a total of [llength $quotes] quotes which is\
                    [string range [expr ([llength $quotes].0/$maxqid) * 100]\
                    0 3]% of all quotes."
    } elseif { [string is integer -strict $arg] } {
      mysqlsel $handle "SELECT quote,name,YEAR(timestamp),\
                        LEFT(MONTHNAME(timestamp), 3),DAYOFMONTH(timestamp),\
                        HOUR(timestamp),MINUTE(timestamp),DAYNAME(timestamp),\
                        CHAR_LENGTH(quote) FROM quotes WHERE qid = $arg"

      mysqlmap $handle { quote name year month day hour min weekday chars } {
        set output "Quote $b$arg$b was added [fix_time $hour $min] $weekday\
                    $day $month $year by $name and is $chars chars long..."
      }
      if { ![info exists output] } {
        set output "ERROR: quote $arg not found."
      }
    } else {
      set quote_max  [mysqlsel $handle "SELECT MAX(qid) FROM quotes" -flatlist]
      set quote_hits [mysqlsel $handle "SELECT COUNT(*) FROM quotes\
                                        WHERE quote LIKE '%$arg%'" -flatlist]
      if { $quote_hits == 0 } {
        set output "$nick: no match for \"$arg\""
      } else {
        set output "$quote_hits of $quote_max quotes contains \"$arg\" which is\
                    $b[string range [expr ($quote_hits.0 / $quote_max) * 100]\
                    0 3]%$b of all quotes."
      }
    }
  } else {
    set output "ERROR: You have to specify a quote, a string or -nick <nick>"
  }
  putcmdlog "#$hand# quote info on $chan"
  fmt_putchan $chan $output
  return 0
}

#------------------------------------------------------------------------------
# Find all quotes and print index to channel.
#------------------------------------------------------------------------------
proc list_quote { nick uhost hand chan args } {
  global handle mysqldb
  if { ![active_chan $chan] } { return 0 }

  mysqluse $handle $mysqldb

  set arg  [strip_bad [lindex [split [lindex $args 0] " "] 0]]
  set arg2 [strip_bad [lindex [split [lindex $args 0] " "] 1]]

  if { $arg != "" } { 
    if { ![string compare -nocase $arg "-nick"] } {
      set arg $arg2
      set quotehits [mysqlsel $handle "SELECT qid FROM quotes\
                                       WHERE name = '$arg'" -list]
    } else {
      set arg [strip_bad [lindex $args 0]]
      set quotehits [mysqlsel $handle "SELECT qid FROM quotes\
                                       WHERE quote LIKE '%$arg%'" -list]
    }
    if { $quotehits == "" } {
      set output "$nick: no match for \"$arg\""
    } else {
      set output "[llength $quotehits] matching quotes: [concat $quotehits]"
    }
  } else {
    set output "ERROR: You have to specify a search string or -nick <nick>"
  }
  putcmdlog "#$hand# qlisting on $chan"
  fmt_putchan $chan $output
  return 0
}

#------------------------------------------------------------------------------
# Find all quotes from a certain date and print index to channel.
#------------------------------------------------------------------------------
proc date_quote { nick uhost hand chan args } {
  global handle mysqldb b
  if { ![active_chan $chan] } { return 0 }

  set arg [strip_bad [lindex $args 0]]
  mysqluse $handle $mysqldb

  if { $arg != "" } { 
    set quotehits [mysqlsel $handle "SELECT qid FROM quotes\
                                     WHERE LEFT(timestamp,8) = '$arg'" -list]
    if { $quotehits == "" } {
      set output "$nick: no quotes added on $arg"
    } else {
      set output "$b[llength $quotehits]$b matching quotes: [concat $quotehits]"
    }
  } else {
    set output "ERROR: You have to specify a date (YYYYMMDD)"
  }
  fmt_putchan $chan $output
  return 0
}

#------------------------------------------------------------------------------
# Show quote status on channel.
#------------------------------------------------------------------------------
proc stat_quote { nick uhost hand chan args } {
  global handle mysqldb
  if { ![active_chan $chan] } { return 0 }

  mysqluse $handle $mysqldb

  set quotes [mysqlsel $handle "SELECT COUNT(qid) FROM quotes" -flatlist]
  set avg [string range [mysqlsel $handle "SELECT AVG(CHAR_LENGTH(quote))\
                                             FROM quotes" -flatlist] 0 5]
  set last_quote [lindex [mysqlsel $handle "SELECT name FROM quotes\
                    WHERE qid = $quotes" -flatlist] 0]
  set output "$nick: Quotes stored: $quotes. Last quote was added\
                by $last_quote. Average length of all quotes: $avg chars."

  putchan $chan $output
  return 0
}

#------------------------------------------------------------------------------
# Show top quoters on channel.
#------------------------------------------------------------------------------
proc top_quote { nick uhost hand chan args } {
  global handle mysqldb b qtop
  if { ![active_chan $chan] } { return 0 }

  set arg [strip_bad [lindex $args 0]]
  mysqluse $handle $mysqldb

  if { ![string is integer -strict $arg] || $arg < 1 } { set arg $qtop }
  putcmdlog "#$hand# showing top $arg quoters on $chan"

  set output "Top $arg quoters: "
  set users [lrange [lsort -decreasing -dictionary -index 1\
             [mysqlsel $handle "SELECT name, COUNT(*) from quotes GROUP BY\
             name" -list]] 0 [incr arg -1]]
  foreach user $users {
    set output "$output $b[lindex $user 0]$b: [lindex $user 1] "
  }

  fmt_putchan $chan $output
  return 0
}

#------------------------------------------------------------------------------
# Print last quote to channel.
#------------------------------------------------------------------------------
proc last_quote { nick uhost hand chan args } {
  global handle mysqldb
  if { ![active_chan $chan] } { return 0 }

  set arg  [strip_bad [lindex [split [lindex $args 0] " "] 0]]
  set arg2 [strip_bad [lindex [split [lindex $args 0] " "] 1]]
  mysqluse $handle $mysqldb

  if { $arg == "" } {
    putcmdlog "#$nick# execing qlast on $chan"
    set qid "-0"
  } else {
    if { [string is integer -strict $arg2] } {
      set num_last $arg2
    } else {
      set num_last 5
    }
    set qid ""
    putcmdlog "#$nick# execing qlast on $chan for: $arg"
    catch { set qid [lindex [mysqlsel $handle "SELECT MAX(qid) FROM quotes\
                     WHERE name = '$arg'" -flatlist] 0] }
  }

  if { $qid == "" } {
    putchan $chan "ERROR: User $arg does not exist"
  } else {
    if { [info exists num_last] } {
      set output "Last $num_last quotes by $arg: "
      set output $output[mysqlsel $handle "SELECT qid FROM quotes WHERE name = '$arg' ORDER BY qid DESC LIMIT $num_last" -flatlist]
      fmt_putchan $chan $output
    }
    get_quote $nick $uhost $hand $chan $qid
  }
  return 0
}

#------------------------------------------------------------------------------
# Show help on channel.
#------------------------------------------------------------------------------
proc help_quote { nick uhost hand chan args } {
  global homepage
  if { ![active_chan $chan] } { return 0 }

  putchan $chan "Commands available: !q(uote) !qput !qstat !qtop !qfind !qlist\
                 !qdate !qinfo !qdel"
  putchan $chan "See $homepage for more info and some stats"
  return 0
}

#------------------------------------------------------------------------------
# Show that we're up and running.
#------------------------------------------------------------------------------
putlog "Quote v2.4 by km@grogg.org loaded"

#------------------------------------------------------------------------------
# EOF
#------------------------------------------------------------------------------
